@@ -157,6 +157,8 @@ class TaskMessageContractor(Contractor):
     _UPLOAD_TASK_ACHIEVE = 'TaskResultUpload'
     _TASK_FINISH = 'TaskFinish'
 
+    _TITLE_ALL = (_TASK_INSIDE, _UPLOAD_TASK_ACHIEVE, _TASK_FINISH)
+
     def __init__(self, task_id: str) -> None:
         super().__init__()
         self.task_id = task_id
@@ -193,7 +195,7 @@ class TaskMessageContractor(Contractor):
             isinstance(event, ContractEvent)
         ), f'"event" should be a contract event object: {event}'
         assert (
-            message_title and message_title in (self._TASK_INSIDE, self._TASK_FINISH)
+            message_title and message_title in self._TITLE_ALL
         ), f'invalid message_title: {message_title}'
 
         event.validate()
@@ -287,7 +289,7 @@ class TaskMessageContractor(Contractor):
                 else:
                     logger.warn(f'failed to parse a contract to an event: {_contract}')
             else:
-                time.sleep(1)
+                time.sleep(0.1)
             continue
 
     @retry()
