@@ -6,7 +6,7 @@ from typing import Dict, Tuple
 import numpy as np
 import pandas as pd
 import torch
-import torch.backends.cudnn as cudnn
+# import torch.backends.cudnn as cudnn
 import torch.nn.functional as F
 import torch.utils.model_zoo as model_zoo
 from imblearn.metrics import sensitivity_score, specificity_score
@@ -338,6 +338,9 @@ class DemoFedIRM(FedAvgScheduler):
     def __init__(self,
                  min_clients: int,
                  max_clients: int,
+                 root_path: str,
+                 csv_file_train: str,
+                 csv_file_test: str,
                  name: str = None,
                  max_rounds: int = 0,
                  merge_epochs: int = 1,
@@ -351,9 +354,6 @@ class DemoFedIRM(FedAvgScheduler):
                  seed: int = 1337,
                  label_uncertainty: str = 'U-Ones',
                  drop_rate: float = 0.2,
-                 root_path: str = f"{CURRENT_DIR}/data/gtr21/ISIN-2018/train_image_224/",
-                 csv_file_train: str = f"{CURRENT_DIR}/data/train.csv",
-                 csv_file_test: str = f"{CURRENT_DIR}/data/test.csv",
                  batch_size: int = 16,
                  base_lr: float = 2e-4):
         super().__init__(min_clients=min_clients,
@@ -378,11 +378,11 @@ class DemoFedIRM(FedAvgScheduler):
         self.base_lr = base_lr
 
         self.is_cuda = torch.cuda.is_available()
-        torch.backends.cudnn.enabled = self.is_cuda
-        if self.is_cuda and self.is_deterministic:
-            cudnn.benchmark = False
-            cudnn.deterministic = True
-            torch.cuda.manual_seed(self.seed)
+        # torch.backends.cudnn.enabled = self.is_cuda
+        # if self.is_cuda and self.is_deterministic:
+        #     cudnn.benchmark = False
+        #     cudnn.deterministic = True
+        #     torch.cuda.manual_seed(self.seed)
 
         self.epochs = 1
 
