@@ -46,8 +46,8 @@ class RightHalfMNIST(torchvision.datasets.MNIST):
         return self.root
 
     def __getitem__(self, index: int) -> Tuple[Any, Any]:
-        img, target = super().__getitem__(index)
-        return self._erase_left(img), target
+        img, _ = super().__getitem__(index)
+        return self._erase_left(img)
 
     def _erase_left(self, _image: torch.Tensor) -> torch.Tensor:
         return _image[:, :, 14:]
@@ -137,7 +137,7 @@ class DemoHeteroHost(HeteroNNHostScheduler):
 
     def load_local_ids(self) -> List[str]:
         train_ids = [str(_id) for _id in self.train_loader.sampler]
-        test_ids = [str(-1 - _id) for _id in self.train_loader.sampler]
+        test_ids = [str(-1 - _id) for _id in self.test_loader.sampler]
         return train_ids + test_ids
 
     def make_feature_model(self) -> nn.Module:
@@ -276,7 +276,7 @@ class DemoHeteroCollaborater(HeteroNNCollaboraterScheduler):
 
     def load_local_ids(self) -> List[str]:
         train_ids = [str(_id) for _id in self.train_loader.sampler]
-        test_ids = [str(-1 - _id) for _id in self.train_loader.sampler]
+        test_ids = [str(-1 - _id) for _id in self.test_loader.sampler]
         return train_ids + test_ids
 
     def make_feature_model(self) -> nn.Module:
@@ -295,14 +295,14 @@ class DemoHeteroCollaborater(HeteroNNCollaboraterScheduler):
                               feature_model: nn.Module,
                               train_ids: List[str]) -> torch.Tensor:
         assert len(train_ids) == 60000, 'Some train samples lost.'
-        for _data, _ in self.train_loader:
+        for _data in self.train_loader:
             yield feature_model(_data)
 
     def iterate_test_feature(self,
                              feature_model: nn.Module,
                              test_ids: List[str]) -> torch.Tensor:
         assert len(test_ids) == 10000, 'Some test samples lost.'
-        for _data, _ in self.test_loader:
+        for _data in self.test_loader:
             yield feature_model(_data)
 
     # replace data channel ports used for debuging
@@ -332,7 +332,7 @@ def get_host():
     return DemoHeteroHost(feature_key='demo_host',
                           project_layer_config=project_layer_config,
                           project_layer_lr=0.01,
-                          batch_size=64,
+                          batch_size=128,
                           data_dir=_DATA_DIR,
                           name='demo_hetero_host',
                           max_rounds=5,
@@ -344,7 +344,7 @@ def get_host():
 def get_collaborater():
     return DemoHeteroCollaborater(feature_key='demo_collaborater',
                                   project_layer_lr=0.01,
-                                  batch_size=64,
+                                  batch_size=128,
                                   data_dir=_DATA_DIR,
                                   name='demo_hetero_collaborater',
                                   data_channel_timeout=(200, 200))
