@@ -14,9 +14,10 @@ from typing import Dict, Tuple
 
 import torch
 
+from ..scheduler import ConfigError
 from .contractor import UploadTrainingResultsEvent
-from . import AggregationError, ConfigError, FedAvgScheduler
 from .dp_contractor import DPCheckinEvent, DPFedAvgContractor
+from .fed_avg import AggregationError, FedAvgScheduler
 
 __all__ = ['DPFedAvgScheduler']
 
