@@ -5,15 +5,12 @@ Reference: https://arxiv.org/pdf/2007.06849.pdf
 
 import io
 import os
-import random
+from tempfile import TemporaryFile
 import time
 from abc import ABC, ABCMeta, abstractmethod
-from dataclasses import dataclass
-from tempfile import TemporaryFile
-from typing import Dict, List, Set, Tuple, Union
+from typing import Dict, List, Set, Tuple
 from zipfile import ZipFile
 
-import tenseal as ts
 import torch
 import torch.nn as nn
 import torch.optim as optim
@@ -21,163 +18,15 @@ import torch.optim as optim
 from .. import logger
 from ..data_channel import GRPCDataChannel
 from ..scheduler import ConfigError, Scheduler, TaskFailed
-from .contractor import (CheckinEvent, CollaboraterCompleteEvent, HeteroNNContractor, SendFeatureEvent,
-                         SendNoisedPorjectionEvent, SendNoisedWGradEvent,
+from .contractor import (CheckinEvent, CollaboraterCompleteEvent,
+                         HeteroNNContractor, SendFeatureEvent,
                          SendTestFeatureEvent, SyncStateResponseEvent)
 from .psi import RSAPSIInitiatorScheduler
 
 __all__ = ['HeteroNNHostScheduler']
 
-_HE_Tensor = Union[torch.Tensor, ts.CKKSTensor]
 _FEATURE_KEY = str
 
-_ENCODING = 'utf-8'
-_LEN_BYTES = 4
-_BYTES_ORDER = 'big'
-
-
-@dataclass
-class _CKKSTensorWrapper:
-
-    feature_key: str
-    cipher: bytes
-    context: bytes = None
-
-    def to_bytes(self) -> bytes:
-        """Encode to bytes data."""
-        assert (
-            self.feature_key and isinstance(self.feature_key, str)
-        ), f'Invalid feature key: {self.feature_key}.'
-        assert self.cipher and isinstance(self.cipher, bytes), f'Invalid cipher: {self.cipher}.'
-        assert (
-            not self.context or isinstance(self.context, bytes)
-        ), f'Invalid context: {self.context}.'
-
-        data_stream = b''
-        key_bytes = self.feature_key.encode(encoding=_ENCODING)
-        key_len = len(key_bytes)
-        data_stream += key_len.to_bytes(_LEN_BYTES, _BYTES_ORDER) + key_bytes
-        cipher_len = len(self.cipher)
-        data_stream += cipher_len.to_bytes(_LEN_BYTES, _BYTES_ORDER) + self.cipher
-        data_stream += self.context if self.context else b''
-        return data_stream
-
-    @classmethod
-    def from_bytes(self, data: bytes) -> '_CKKSTensorWrapper':
-        """Initialize a _FeatureCipher object from bytes data."""
-        archor = 0
-        assert len(data) > archor + _LEN_BYTES, 'Invalid feature key length data.'
-        key_len = int.from_bytes(data[archor:(archor + _LEN_BYTES)], _BYTES_ORDER)
-        archor += _LEN_BYTES
-        assert len(data) > archor + key_len, 'Invalid feature key data.'
-        feature_key = data[archor:(archor + key_len)].decode(encoding=_ENCODING)
-        archor += key_len
-        assert len(data) > archor + _LEN_BYTES, 'Invalid cipher length data.'
-        cipher_len = int.from_bytes(data[archor:(archor + _LEN_BYTES)], _BYTES_ORDER)
-        archor += _LEN_BYTES
-        assert len(data) >= archor + cipher_len, 'Invalid cipher data.'
-        cipher = data[archor:(archor + cipher_len)]
-        archor += cipher_len
-        context = data[archor:] or None
-        return _CKKSTensorWrapper(feature_key=feature_key, cipher=cipher, context=context)
-
-
-@dataclass
-class _NoisedWGradWrapper:
-
-    feature_key: str
-    noised_grad: torch.Tensor
-    cipher_epsilon_acc: bytes
-    context: bytes
-
-    def to_bytes(self) -> bytes:
-        """Encode to bytes data."""
-        assert (
-            self.feature_key and isinstance(self.feature_key, str)
-        ), f'Invalid feature key: {self.feature_key}.'
-        assert (
-            self.noised_grad is not None and isinstance(self.noised_grad, torch.Tensor)
-        ), f'Invalid noised grad: {self.noised_grad}.'
-        assert (
-            self.cipher_epsilon_acc and isinstance(self.cipher_epsilon_acc, bytes)
-        ), f'Invalid cipher epsilon acc: {self.cipher_epsilon_acc}.'
-        assert (
-            self.context and isinstance(self.context, bytes)
-        ), f'Invalid context: {self.context}.'
-
-        data_stream = b''
-        key_bytes = self.feature_key.encode(encoding=_ENCODING)
-        key_len = len(key_bytes)
-        data_stream += key_len.to_bytes(_LEN_BYTES, _BYTES_ORDER) + key_bytes
-        with TemporaryFile() as tf:
-            torch.save(self.noised_grad, tf)
-            tf.seek(0)
-            grad_bytes = tf.read()
-            grad_len = len(grad_bytes)
-            data_stream += grad_len.to_bytes(_LEN_BYTES, _BYTES_ORDER) + grad_bytes
-        cipher_len = len(self.cipher_epsilon_acc)
-        data_stream += cipher_len.to_bytes(_LEN_BYTES, _BYTES_ORDER) + self.cipher_epsilon_acc
-        data_stream += self.context
-        return data_stream
-
-    @classmethod
-    def from_bytes(self, data: bytes) -> '_NoisedWGradWrapper':
-        """Initialize a _NoisedWGradWrapper object from bytes data."""
-        archor = 0
-        assert len(data) > archor + _LEN_BYTES, 'Invalid feature key length data.'
-        key_len = int.from_bytes(data[archor:(archor + _LEN_BYTES)], _BYTES_ORDER)
-        archor += _LEN_BYTES
-        assert len(data) > archor + key_len, 'Invalid feature key data.'
-        feature_key = data[archor:(archor + key_len)].decode(encoding=_ENCODING)
-        archor += key_len
-        assert len(data) > archor + _LEN_BYTES, 'Invalid grad length data.'
-        grad_len = int.from_bytes(data[archor:(archor + _LEN_BYTES)], _BYTES_ORDER)
-        archor += _LEN_BYTES
-        assert len(data) >= archor + grad_len, 'Invalid grad data.'
-        grad_bytes = data[archor:(archor + grad_len)]
-        buffer = io.BytesIO(grad_bytes)
-        noised_grad = torch.load(buffer)
-        archor += grad_len
-        assert len(data) > archor + _LEN_BYTES, 'Invalid cipher length data.'
-        cipher_len = int.from_bytes(data[archor:(archor + _LEN_BYTES)], _BYTES_ORDER)
-        archor += _LEN_BYTES
-        assert len(data) >= archor + cipher_len, 'Invalid cipher data.'
-        cipher_epsilon_acc = data[archor:(archor + cipher_len)]
-        archor += cipher_len
-        assert len(data) > archor, 'Invalid context data.'
-        context = data[archor:]
-        return _NoisedWGradWrapper(feature_key=feature_key,
-                                   noised_grad=noised_grad,
-                                   cipher_epsilon_acc=cipher_epsilon_acc,
-                                   context=context)
-
-
-class _ProjectLayer(nn.Module):
-    """An interactive layer used for fusion features with privacy preserving."""
-
-    def __init__(self, project_config: List[Tuple[str, int, int]]) -> None:
-        super().__init__()
-        for _key, _in, _out in project_config:
-            if not _key or not isinstance(_key, str):
-                raise ConfigError(f'Invalid key in project rule: {project_config}')
-            if not _in or not isinstance(_in, int) or _in < 1:
-                raise ConfigError(f'Invalid input dimension in project rule: {project_config}')
-            if not _out or not isinstance(_out, int) or _out < 1:
-                raise ConfigError(f'Invalid output dimension in project rule: {project_config}')
-
-        for _key, _in, _out in project_config:
-            self.add_module(name=_key, module=nn.Linear(_in, _out, bias=False))
-
-    def forward(self, input_map: Dict[str, _HE_Tensor]) -> Dict[str, _HE_Tensor]:
-        output_map = {}
-        for _key, _input in input_map.items():
-            if isinstance(_input, ts.CKKSTensor):
-                W_tensor = self.__getattr__(_key).weight
-                output_map[_key] = _input.mm(W_tensor.T.tolist())
-            else:
-                output_map[_key] = self.__getattr__(_key)(_input)
-        return output_map
-
 
 class _SimplifiedOptimizer(ABC):
     """A simplified optimizer tool to facilitate update parameters."""
@@ -322,12 +171,8 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
     """Schedule the process of the host in a hetero_nn task."""
 
     _WAITING_FOR_FEATURES = 'wait_4_feature'
-    _DISTRIBUTING_CIPHER_PROJECTION = 'distribute_cipher_proj'
-    _COLLECTING_NOISED_PROJECTION = 'collect_proj'
     _GETTING_GRAD = 'calc_loss'
-    _DISTRIBUTING_CIPHER_W_GRAD = 'distribute_w_grad'
-    _COLLECTING_NOISED_W_GRAD = 'collect_w_grad'
-    _DISTRIBUTING_CIPHER_FEATURE_GRAD = 'distribute_feature_grad'
+    _DISTRIBUTING_FEATURE_GRAD = 'distribute_grad'
 
     def __init__(self,
                  feature_key: str,
@@ -393,7 +238,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
 
         self._validate_config()
 
-        self.project_layer
         self.infer_model
         self.infer_optimizer
 
@@ -403,12 +247,9 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         self._example_project_input = None
         self._example_infer_input = None
 
-        self._alpha_map: Dict[str, Dict[_FEATURE_KEY, _HE_Tensor]] = {}
-        self._feature_projection_map: Dict[_FEATURE_KEY, _HE_Tensor] = {}
+        self._alpha_map: Dict[str, Dict[_FEATURE_KEY, torch.Tensor]] = {}
         self._feature_fusion_map: Dict[_FEATURE_KEY, torch.Tensor] = {}
-        self._noised_w_grad_map: Dict[_FEATURE_KEY, _NoisedWGradWrapper] = {}
-        self._batched_test_features: List[List[Dict[_FEATURE_KEY, _HE_Tensor]]] = []
-        self._epsilon_host = None
+        self._batched_test_features: List[List[Dict[_FEATURE_KEY, torch.Tensor]]] = []
 
     def _validate_config(self):
         if not self.feature_key or not isinstance(self.feature_key, str):
@@ -432,53 +273,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         ):
             raise ConfigError(f'Invalid project layer learning rate: {self.project_layer_lr}.')
 
-    @property
-    def project_layer(self) -> nn.Module:
-        """Return a model object to project features."""
-        if not hasattr(self, '_project_layer'):
-            def get_example_input(module, input, output):
-                if self._example_project_input is None:
-                    self._example_project_input = input
-
-            self._project_layer = _ProjectLayer(project_config=self.project_layer_config)
-            self._project_layer.register_forward_hook(get_example_input)
-
-        return self._project_layer
-
-    @property
-    def project_optimizer(self) -> _SimplifiedOptimizer:
-        """Return a proxy optimizer to facilitate updating parameters of the project layer."""
-
-        class _ProjectOptimizerImpl(_SimplifiedOptimizer):
-
-            def __init__(self, host_obj: HeteroNNHostScheduler) -> None:
-                super().__init__()
-                self.host_obj = host_obj
-
-            def zero_grad(self):
-                if self.host_obj.project_layer is not None:
-                    for _param in self.host_obj.project_layer.parameters():
-                        if _param.grad is not None:
-                            _param.grad.zero_()
-                if self.host_obj._feature_fusion_map is not None:
-                    for _proj in self.host_obj._feature_fusion_map.values():
-                        if _proj.grad is not None:
-                            _proj.grad.zero_()
-
-            def step(self):
-                self.host_obj._switch_status(self.host_obj._DISTRIBUTING_CIPHER_W_GRAD)
-                self.host_obj._distribute_cipher_w_grad()
-                self.host_obj._switch_status(self.host_obj._COLLECTING_NOISED_W_GRAD)
-                self.host_obj._collect_noised_w_grad()
-                self.host_obj._switch_status(self.host_obj._DISTRIBUTING_CIPHER_FEATURE_GRAD)
-                self.host_obj._distribute_cipher_feature_grad()
-                self.host_obj._switch_status(self.host_obj._UPDATING)
-                self.host_obj._update_project_layer_weight()
-
-        if not hasattr(self, '_project_optimizer'):
-            self._project_optimizer = _ProjectOptimizerImpl(host_obj=self)
-        return self._project_optimizer
-
     @abstractmethod
     def make_infer_model(self) -> nn.Module:
         """Return a model object to infer business results."""
@@ -523,12 +317,10 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
 
             def zero_grad(self):
                 self.host_obj.infer_optimizer.zero_grad()
-                self.host_obj.project_optimizer.zero_grad()
                 self.host_obj.feature_optimizer.zero_grad()
 
             def step(self):
                 self.host_obj.infer_optimizer.step()
-                self.host_obj.project_optimizer.step()
                 self.host_obj.feature_optimizer.step()
 
         if not hasattr(self, '_optimizer'):
@@ -623,9 +415,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             err_msg += f'Got a {type(self.infer_optimizer)} object.'
             raise ConfigError(err_msg)
 
-        if self.project_layer is None:
-            raise TaskFailed('Failed to initialize the project layer.')
-
         if not self._partners:
             raise TaskFailed('No partners.')
 
@@ -653,7 +442,7 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         """Is the max rounds configuration reached."""
         return self._round >= self.max_rounds
 
-    def _validate_feature_dict(self, features: Dict[str, ts.CKKSTensor]):
+    def _validate_feature_dict(self, features: Dict[str, torch.Tensor]):
         """Validate feature format."""
         if not features or not isinstance(features, dict) or len(features) != 1:
             self.push_log(f'Received invalid features: {features}')
@@ -663,7 +452,7 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         if not _key or not isinstance(_key, str):
             self.push_log(f'Received invalid feature key: {_key}')
             raise TaskFailed('Invalid feature type. It must contain a keyword of string.')
-        if _val is None or not isinstance(_val, ts.CKKSTensor) or len(_val.shape) != 2:
+        if _val is None or not isinstance(_val, torch.Tensor) or _val.dim() != 2:
             self.push_log(f'Received invalid feature value: {_val}')
             raise TaskFailed('Invalid feature type. Its value must be a tensor of two dimension.')
 
@@ -761,7 +550,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
     def _run_a_round(self):
         self._start_round()
         self.infer_model.train()
-        self.project_layer.train()
         self.feature_model.train()
         time.sleep(1)  # TODO remove me
         for _feature_batch, _labels in self.iterate_train_feature(
@@ -771,15 +559,10 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             self._local_features = _feature_batch
             self._switch_status(self._WAITING_FOR_FEATURES)
             self._collect_features()
-            self._switch_status(self._PROJECTING)
-            self._make_projection()
-            self._switch_status(self._DISTRIBUTING_CIPHER_PROJECTION)
-            self._distribute_cipher_projection()
-            self._switch_status(self._COLLECTING_NOISED_PROJECTION)
-            self._collect_noised_projection()
-
             self._switch_status(self._GETTING_GRAD)
             self.train_a_batch(self._feature_fusion_map, _labels)
+            self._switch_status(self._DISTRIBUTING_FEATURE_GRAD)
+            self._distribute_feature_grad()
 
         self._switch_status(self._PERSISTING)
         self._save_model()
@@ -794,161 +577,52 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         self.contractor.start_round(round=self._round)
         self.push_log(f'Calculation of round {self._round} is started.')
 
-    def _collect_features(self) -> Dict[str, _HE_Tensor]:
+    def _collect_features(self) -> Dict[str, torch.Tensor]:
         """Collect all input features from all partners."""
         self.push_log('Waiting for collecting all features from partners ...')
         self.contractor.notify_ready_for_fusion(self._round)
-        feature_map: Dict[str, Dict[str, _HE_Tensor]] = {
+        feature_map: Dict[str, Dict[str, torch.Tensor]] = {
             self.id: {self.feature_key: self._local_features}
         }
         for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
             if isinstance(_event, SendFeatureEvent):
                 feature_stream = self.data_channel.receive_stream(_event)
-                cipher_feature = _CKKSTensorWrapper.from_bytes(feature_stream)
-                cipher_context = ts.context_from(cipher_feature.context)
-                cipher_tensor = ts.ckks_tensor_from(cipher_context, cipher_feature.cipher)
-                features = {cipher_feature.feature_key: cipher_tensor}
+                buffer = io.BytesIO(feature_stream)
+                features = torch.load(buffer)
                 self._validate_feature_dict(features)
                 feature_map[_event.source] = features
                 if len(feature_map) == len(self._partners) + 1:  # plus self
                     self._alpha_map = feature_map
+                    features = dict(feature_dict.copy().popitem()
+                                    for feature_dict in self._alpha_map.values())
+                    self._feature_fusion_map = features
                     return
         # timeout
         raise TaskFailed('Failed to collect all features.')
 
-    def _make_projection(self):
-        """Fuse features and get cipher feature projection."""
-        self.push_log('Calculating features projection ...')
-        features = dict(feature_dict.copy().popitem()
-                        for feature_dict in self._alpha_map.values())
-        self._feature_projection_map = self.project_layer(features)
-
-    def _distribute_cipher_projection(self):
-        """Distribute cipher feature projection to collaboraters."""
-        self.push_log('Distributing cipher projection to collaboraters ...')
-        self._epsilon_host = random.randint(1, 2**10)
-        for _partner, _cipher_feature in self._alpha_map.items():
+    def _distribute_feature_grad(self):
+        """Distribute feature grad tensors to collaboraters."""
+        self.push_log('Distributing features grad tensors ...')
+        for _partner, _feature_dict in self._alpha_map.items():
             if _partner == self.id:
                 continue
-            feature_key, _ = _cipher_feature.copy().popitem()
-            cipher_tensor: ts.CKKSTensor = self._feature_projection_map[feature_key]
-            cipher_tensor.add_(self._epsilon_host)
-            cipher_projection = _CKKSTensorWrapper(feature_key=feature_key,
-                                                   cipher=cipher_tensor.serialize())
-            self.data_channel.send_stream(source=self.id,
-                                          target=_partner,
-                                          data_stream=cipher_projection.to_bytes(),
-                                          connection_timeout=self.dc_conn_timeout,
-                                          timeout=self.dc_timeout)
-            self.push_log(f'Distributed cipher projection to collaborater ID: {_partner}')
-
-    @torch.no_grad()
-    def _collect_noised_projection(self):
-        """Collect noised but plain projection from collaboraters."""
-        self.push_log('Collecting noised but plain projection from collaboraters ...')
-        self.contractor.notify_ready_for_noised_projection(round=self._round)
-        self._feature_fusion_map = {
-            self.feature_key: self._feature_projection_map[self.feature_key]
-        }
-        for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
-            if isinstance(_event, SendNoisedPorjectionEvent):
-                stream = self.data_channel.receive_stream(_event)
-                buffer = io.BytesIO(stream)
-                proj_dict: Dict[str, torch.Tensor] = torch.load(buffer)
-                feature_key, proj_tensor = proj_dict.copy().popitem()
-                proj_tensor.sub_(self._epsilon_host)
-                proj_tensor.requires_grad_()
-                self._feature_fusion_map[feature_key] = proj_tensor
-                self.push_log(f'Received noised projection from ID: {_event.source}.')
-                if len(self._feature_fusion_map) == len(self._partners) + 1:  # and self
-                    self.push_log('Received all copies of noised projection.')
-                    return
-        # timeout
-        raise TaskFailed('Failed to collect all noised projection.')
-
-    def _distribute_cipher_w_grad(self):
-        self._epsilon_host = random.randint(1, 2**10)
-        self.push_log('Distributing cipher W grad to collaboraters ...')
-        for _partner, feature_dict in self._alpha_map.items():
-            if _partner == self.id:
-                continue
-            feature_key, feature_tensor = feature_dict.copy().popitem()
-            feature_tensor: ts.CKKSTensor
-            proj = self._feature_fusion_map[feature_key]
-            cipher_W_grad: ts.CKKSTensor = feature_tensor.transpose()
-            cipher_W_grad.mm_(proj.grad).add_(self._epsilon_host)
-            self.data_channel.send_stream(source=self.id,
-                                          target=_partner,
-                                          data_stream=cipher_W_grad.serialize(),
-                                          connection_timeout=self.dc_conn_timeout,
-                                          timeout=self.dc_timeout)
-            self.push_log(f'Sent cipher W grad to ID: {_partner}.')
-        self.push_log('Distributed all cipher W grad to collaboraters.')
-
-    def _collect_noised_w_grad(self):
-        """Collect decrypted but noised W grad of project layer from collaboraters."""
-        self.contractor.notify_ready_for_noised_w_grad(round=self._round)
-        self.push_log('Collecting noised W grad from collaboraters ...')
-        self._noised_w_grad_map.clear()
-        for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
-            if isinstance(_event, SendNoisedWGradEvent):
-                stream = self.data_channel.receive_stream(_event)
-                noised_grad = _NoisedWGradWrapper.from_bytes(stream)
-                self._noised_w_grad_map[noised_grad.feature_key] = noised_grad
-                self.push_log(f'Received noised W grad from parter ID: {_event.source}.')
-                if len(self._noised_w_grad_map) == len(self._partners):
-                    break
-        self.push_log('Received all copies of noised W grad.')
-
-    def _distribute_cipher_feature_grad(self):
-        """Send cipher feature grad to collaboraters."""
-        self.push_log('Distributing cipher feature grad to collaboraters ...')
-        for _partner, feature_dict in self._alpha_map.items():
-            if _partner == self.id:
-                continue
-            feature_key, _ = feature_dict.copy().popitem()
-            noised_w_grad = self._noised_w_grad_map[feature_key]
-            he_context = ts.context_from(noised_w_grad.context)
-            cipher_acc = ts.ckks_tensor_from(he_context, noised_w_grad.cipher_epsilon_acc)
-            partner_linear: nn.Linear = self.project_layer.__getattr__(feature_key)
-            partner_W = partner_linear.weight
-            proj_grad = self._feature_fusion_map[feature_key].grad
-            cipher_grad = cipher_acc.broadcast_(partner_W.shape).add_(partner_W.tolist())
-            cipher_grad.mm_(proj_grad.T).transpose_()
-            self.data_channel.send_stream(source=self.id,
-                                          target=_partner,
-                                          data_stream=cipher_grad.serialize(),
-                                          connection_timeout=self.dc_conn_timeout,
-                                          timeout=self.dc_timeout)
-            self.push_log(f'Sent cipher feature grad to ID {_partner}.')
-        self.push_log('Distributed all cipher feature grad to collaboraters.')
-
-    @torch.no_grad()
-    def _update_project_layer_weight(self):
-        """Update parameters of project layer."""
-        self.push_log('Updating parameters of project layer ...')
-
-        # update self owned project linear layer
-        proj_linear: nn.Linear = self.project_layer.__getattr__(self.feature_key)
-        proj_linear.weight.sub_(self.project_layer_lr * proj_linear.weight.grad)
-
-        # update partners' project linear layer
-        for _feature_key, _noised_w_grad in self._noised_w_grad_map.items():
-            proj_linear: nn.Linear = self.project_layer.__getattr__(_feature_key)
-            _noised_tensor = _noised_w_grad.noised_grad
-            _noised_tensor.sub_(self._epsilon_host)
-            proj_linear.weight.sub_(self.project_layer_lr * _noised_tensor)
-
-        self.push_log('Updated parameters of project layer.')
+            _, feature_tensor = _feature_dict.copy().popitem()
+            with TemporaryFile() as tf:
+                torch.save(feature_tensor.grad, tf)
+                tf.seek(0)
+                self.data_channel.send_stream(source=self.id,
+                                              target=_partner,
+                                              data_stream=tf.read(),
+                                              connection_timeout=self.dc_conn_timeout,
+                                              timeout=self.dc_timeout)
+        self.push_log('Distributed all features grad tensors to collaboraters.')
 
     def _save_model(self):
         """Save latest model state."""
         os.makedirs(self._save_path, exist_ok=True)
-        with open(f'{os.path.join(self._save_path, "feature_model.pt")}', 'wb') as f:
+        with open(os.path.join(self._save_path, 'feature_model_ckp.pt'), 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
-        with open(f'{os.path.join(self._save_path, "project_layer.pt")}', 'wb') as f:
-            torch.save(self.project_layer.state_dict(), f)
-        with open(f'{os.path.join(self._save_path, "infer_model.pt")}', 'wb') as f:
+        with open(os.path.join(self._save_path, 'infer_model_ckp.pt'), 'wb') as f:
             torch.save(self.infer_model.state_dict(), f)
         self.push_log('Saved latest parameters locally.')
 
@@ -963,28 +637,25 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             self.push_log('Start a round of test.')
 
             self.feature_model.eval()
-            self.project_layer.eval()
             self.infer_model.eval()
 
             self.contractor.start_test_round(round=self._round)
 
-            batched_feature_projections = []
+            batched_host_features = []
             batched_labels = []
             for _feature_batch, _labels in self.iterate_test_feature(
                 self.feature_model, self.test_ids
             ):
-                self._local_features = _feature_batch
-                self._switch_status(self._WAITING_FOR_FEATURES)
-                self._collect_features()
-                self._switch_status(self._PROJECTING)
-                self._make_projection()
-                self._switch_status(self._DISTRIBUTING_CIPHER_PROJECTION)
-                self._distribute_cipher_projection()
-                self._switch_status(self._COLLECTING_NOISED_PROJECTION)
-                self._collect_noised_projection()
+                batched_host_features.append((self.feature_key, _feature_batch))
                 batched_labels.append(_labels)
-                batched_feature_projections.append(self._feature_fusion_map)
-                self.push_log('Fused a batch of test data features.')
+
+            self._switch_status(self._WAITING_FOR_FEATURES)
+            self._wait_for_testing_features()
+            self._batched_test_features.append(batched_host_features)
+            self._switch_status(self._PROJECTING)
+            batched_feature_projections = [dict(_batch)
+                                           for _batch in zip(*self._batched_test_features)]
+            self.push_log('Fused test data features.')
 
             self.test(batched_feature_projections=batched_feature_projections,
                       batched_labels=batched_labels)
@@ -1003,7 +674,7 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
                 buffer = io.BytesIO(stream)
                 batched_features: dict = torch.load(buffer)
                 _key, _feature_list = batched_features.copy().popitem()
-                self._batched_test_features.append([{_key: _feature_batch}
+                self._batched_test_features.append([(_key, _feature_batch)
                                                     for _feature_batch in _feature_list])
                 self.push_log(f'Received test dataset features from ID: {_event.source}.')
                 if len(self._batched_test_features) == len(self._partners):
@@ -1055,19 +726,15 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
 
         # torch.jit doesn't work with a TemporaryFile
         feature_model_file = os.path.join(self._save_path,
-                                          f'scripted_feature_{self.feature_key}.pt')
+                                          f'feature_model_{self.feature_key}.pt')
         with open(feature_model_file, 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
-        project_layer_file = os.path.join(self._save_path, 'scripted_project_layer.pt')
-        with open(project_layer_file, 'wb') as f:
-            torch.save(self.project_layer.state_dict(), f)
-        infer_model_file = f'{os.path.join(self._save_path, "scripted_infer.pt")}'
+        infer_model_file = f'{os.path.join(self._save_path, "infer_model.pt")}'
         with open(infer_model_file, 'wb') as f:
             torch.save(self.infer_model.state_dict(), f)
         model_file = os.path.join(self._save_path, 'model.zip')
         with ZipFile(model_file, 'w') as model_zip:
             model_zip.write(feature_model_file)
-            model_zip.write(project_layer_file)
             model_zip.write(infer_model_file)
         model_file_path = os.path.abspath(model_file)
 
