@@ -1,34 +1,26 @@
+import io
 import os
 import random
 from abc import abstractmethod
 from tempfile import TemporaryFile
 from typing import List, Tuple
 
-import tenseal as ts
 import torch
 import torch.nn as nn
 import torch.optim as optim
 
 from .. import logger
 from ..scheduler import ConfigError, TaskComplete, TaskFailed
-from .contractor import (CheckinResponseEvent, CloseRoundEvent,
-                         CloseTestRoundEvent, CompleteTaskEvent, FailTaskEvent,
-                         ReadyForFusionEvent, ReadyForNoisedProjectionEvent,
-                         ReadyForNoisedWGradEvent, SendCipherFeatureGradEvent,
-                         SendCipherPorjectionEvent, SendCipherWGradEvent,
-                         StartRoundEvent, StartTestRoundEvent, SyncStateEvent)
-from .hetero_nn import (HeteroNNScheduler, _CKKSTensorWrapper,
-                        _NoisedWGradWrapper)
+from .contractor import (CheckinResponseEvent, CloseRoundEvent, StartTestRoundEvent,
+                         CloseTestRoundEvent, CompleteTaskEvent, FailTaskEvent, SyncStateEvent,
+                         ReadyForFusionEvent, SendFeatureGradEvent, StartRoundEvent)
+from .hetero_nn import (HeteroNNScheduler)
 from .psi import RSAPSICollaboratorScheduler
 
 
 class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
     """Schedule the process of a collaborater in a hetero_nn task."""
 
-    _WAITING_FOR_CIPHER_PROJECTION = 'wait_4_proj'
-    _SENDING_NOISED_PROJECTION = 'send_proj'
-    _WAITING_FOR_W_GRAD = 'wait_4_w_grad'
-    _SENDING_NOISED_W_GRAD = 'send_w_grad'
     _WAITING_FOR_FEATUE_GRAD = 'wait_4_feature_grad'
 
     def __init__(self,
@@ -76,11 +68,6 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
 
         self.host = None
 
-        self._he_context: ts.Context = None
-        self._he_context_serialize: bytes = None
-        self._epsilon_acc: float = None
-        self._proj_tensor: torch.Tensor = None
-        self._noised_w_grad: torch.Tensor = None
         self._feature_grad: torch.Tensor = None
 
     def _validate_config(self):
@@ -143,13 +130,6 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
             self.name = f'guest_{self.task_id}'
         self._save_path = os.path.join('models', self.name)
 
-        self._he_context = ts.context(ts.SCHEME_TYPE.CKKS,
-                                      poly_modulus_degree=2**12,
-                                      coeff_mod_bit_sizes=[21, 20, 20, 21])
-        self._he_context.global_scale = 2**20
-        self._he_context_serialize = self._he_context.serialize()
-        self._epsilon_acc = random.gauss(0, 0.1)
-
         self.push_log(message='Begin to validate local context.')
         self.validate_context()
 
@@ -244,20 +224,12 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
             self.push_log('Featured a batch of data.')
             self._switch_status(self._PROJECTING)
             self._local_features = _batch_features
-            self._send_feature_cipher()
-            self._switch_status(self._WAITING_FOR_CIPHER_PROJECTION)
-            self._wait_for_cipher_projection()
-            self._switch_status(self._SENDING_NOISED_PROJECTION)
-            self._send_noised_projection()
+            self._send_feature()
 
-            self.feature_optimizer.zero_grad()
-            self._switch_status(self._WAITING_FOR_W_GRAD)
-            self._wait_for_w_grad()
-            self._switch_status(self._SENDING_NOISED_W_GRAD)
-            self._send_noised_w_grad()
             self._switch_status(self._WAITING_FOR_FEATUE_GRAD)
             self._wait_for_feature_grad()
             self._switch_status(self._UPDATING)
+            self.feature_optimizer.zero_grad()
             self._local_features.backward(self._feature_grad)
             self.feature_optimizer.step()
 
@@ -285,7 +257,7 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
 
-    def _send_feature_cipher(self):
+    def _send_feature(self):
         """Send local features of a batch of data to the host."""
         self.push_log('Waiting for sending features ...')
         for _event in self.contractor.contract_events():
@@ -298,102 +270,31 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
                 raise TaskFailed('Aborted by host.')
 
         self.push_log('Begin to send features.')
-        cipher_tensor = ts.ckks_tensor(self._he_context, self._local_features.detach())
-        cipher_feature = _CKKSTensorWrapper(feature_key=self.feature_key,
-                                            cipher=cipher_tensor.serialize(),
-                                            context=self._he_context_serialize)
-        self.data_channel.send_stream(source=self.id,
-                                      target=self.host,
-                                      data_stream=cipher_feature.to_bytes(),
-                                      connection_timeout=self.dc_conn_timeout,
-                                      timeout=self.dc_timeout)
+        with TemporaryFile() as tf:
+            torch.save({self.feature_key: self._local_features}, tf)
+            tf.seek(0)
+            self.data_channel.send_stream(source=self.id,
+                                          target=self.host,
+                                          data_stream=tf.read(),
+                                          connection_timeout=self.dc_conn_timeout,
+                                          timeout=self.dc_timeout)
         self.push_log('Sending features complete.')
 
-    def _wait_for_cipher_projection(self):
-        """Wait for receiving cipher projection of the self owned features."""
-        self.push_log('Waiting for receiving cipher projection ...')
-        for _event in self.contractor.contract_events():
-            if isinstance(_event, SendCipherPorjectionEvent):
-                stream = self.data_channel.receive_stream(_event)
-                cipher_projection = _CKKSTensorWrapper.from_bytes(stream)
-                cipher_tensor = ts.ckks_tensor_from(context=self._he_context,
-                                                    data=cipher_projection.cipher)
-                projection = cipher_tensor.decrypt()
-                self._proj_tensor = torch.tensor(projection.tolist())
-                self._proj_tensor.add_(self._local_features * self._epsilon_acc)
-                self.push_log('Received cipher projection.')
-                return
-            elif isinstance(_event, FailTaskEvent):
-                raise TaskFailed('Aborted by host.')
-
-    def _send_noised_projection(self):
-        """Send noised projection to the host for fusion."""
-        self.push_log('Waiting for sending noised projection ...')
-        for _event in self.contractor.contract_events():
-            if isinstance(_event, ReadyForNoisedProjectionEvent):
-                with TemporaryFile() as tf:
-                    torch.save({self.feature_key: self._proj_tensor}, tf)
-                    tf.seek(0)
-                    self.data_channel.send_stream(source=self.id,
-                                                  target=self.host,
-                                                  data_stream=tf.read(),
-                                                  connection_timeout=self.dc_conn_timeout,
-                                                  timeout=self.dc_timeout)
-                return
-            elif isinstance(_event, FailTaskEvent):
-                raise TaskFailed('Aborted by host.')
-
-    def _wait_for_w_grad(self):
-        """Wait for cipher grad of project layer W to decrypt."""
-        self.push_log('Waiting for cipher grad of project layer W ...')
-        for _event in self.contractor.contract_events():
-            if isinstance(_event, SendCipherWGradEvent):
-                stream = self.data_channel.receive_stream(_event)
-                cipher_grad = ts.ckks_tensor_from(self._he_context, stream)
-                # sample from (-0.005, 0.015) to decrease the overrall speed of
-                # accumulation, thus facilitate convergence
-                epsilon_collab = random.uniform(-0.005, 0.015)
-                self._noised_w_grad = torch.tensor(cipher_grad.decrypt().tolist())
-                self._noised_w_grad.add_(epsilon_collab / self.project_layer_lr)
-                self._epsilon_acc += epsilon_collab
-                break
-            elif isinstance(_event, FailTaskEvent):
-                raise TaskFailed('Aborted by host.')
-        self.push_log('Received and decrypted cipher grad of project layer W.')
-
-    def _send_noised_w_grad(self):
-        """Send decrypted grad of project layer W to the host."""
-        self.push_log('Sending noised grad of project layer W to host ...')
-        for _event in self.contractor.contract_events():
-            cipher_acc = ts.ckks_tensor(self._he_context, [self._epsilon_acc])
-            if isinstance(_event, ReadyForNoisedWGradEvent):
-                noised_w_grad = _NoisedWGradWrapper(feature_key=self.feature_key,
-                                                    noised_grad=self._noised_w_grad,
-                                                    cipher_epsilon_acc=cipher_acc.serialize(),
-                                                    context=self._he_context_serialize)
-                self.data_channel.send_stream(source=self.id,
-                                              target=self.host,
-                                              data_stream=noised_w_grad.to_bytes(),
-                                              connection_timeout=self.dc_conn_timeout,
-                                              timeout=self.dc_timeout)
-                break
-        self.push_log('Sent noised grad of project layer W to host.')
-
     def _wait_for_feature_grad(self):
         """Wait for cipher grad of feature model output."""
         self.push_log('Waiting for cipher grad of feature model output ...')
         for _event in self.contractor.contract_events():
-            if isinstance(_event, SendCipherFeatureGradEvent):
+            if isinstance(_event, SendFeatureGradEvent):
                 stream = self.data_channel.receive_stream(_event)
-                cipher_grad = ts.ckks_tensor_from(self._he_context, stream)
-                self._feature_grad = torch.tensor(cipher_grad.decrypt().tolist())
+                buffer = io.BytesIO(stream)
+                self._feature_grad = torch.load(buffer)
                 break
         self.push_log('Received and decrypted cipher grad of feature model output.')
 
     def _save_model(self):
         """Save latest model state."""
         os.makedirs(self._save_path, exist_ok=True)
-        with open(f'{os.path.join(self._save_path, "feature_model.pt")}', 'wb') as f:
+        with open(f'{os.path.join(self._save_path, "feature_model_ckp.pt")}', 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
         self.push_log('Saved latest parameters locally.')
 
@@ -407,16 +308,19 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
                 ), f'Lost synchronization, host: {_event.round}; local: {self._round}.'
 
                 self.feature_model.eval()
-                for _batch_features in self.iterate_test_feature(self.feature_model,
-                                                                 self.test_ids):
-                    self._switch_status(self._PROJECTING)
-                    self._local_features = _batch_features
-                    self._send_feature_cipher()
-                    self._switch_status(self._WAITING_FOR_CIPHER_PROJECTION)
-                    self._wait_for_cipher_projection()
-                    self._switch_status(self._SENDING_NOISED_PROJECTION)
-                    self._send_noised_projection()
-                    self.push_log('Fused a batch of test data features.')
+                features = {
+                    self.feature_key: list(self.iterate_test_feature(self.feature_model,
+                                                                     self.test_ids))
+                }
+                with TemporaryFile() as tf:
+                    torch.save(features, tf)
+                    tf.seek(0)
+                    self.data_channel.send_stream(source=self.id,
+                                                  target=self.host,
+                                                  data_stream=tf.read(),
+                                                  connection_timeout=self.dc_conn_timeout,
+                                                  timeout=self.dc_timeout)
+                self.push_log('Sent all batches of feature to the host.')
 
             elif isinstance(_event, CloseTestRoundEvent):
                 self.push_log('Skipped or closed a round of testing.')
@@ -458,7 +362,7 @@ class HeteroNNCollaboraterScheduler(HeteroNNScheduler):
         self.push_log('Generating task achievement files ...')
         # torch.jit doesn't work with a TemporaryFile
         feature_model_file = os.path.join(self._save_path,
-                                          f'scripted_feature_{self.feature_key}.pt')
+                                          f'feature_model_{self.feature_key}.pt')
         with open(feature_model_file, 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
         model_file_path = os.path.abspath(feature_model_file)
